@@ -5,7 +5,7 @@ use Carp qw/croak confess carp/;
 
 BEGIN {
   $Math::Prime::Util::GMP::AUTHORITY = 'cpan:DANAJ';
-  $Math::Prime::Util::GMP::VERSION = '0.04';
+  $Math::Prime::Util::GMP::VERSION = '0.05';
 }
 
 # parent is cleaner, and in the Perl 5.10.1 / 5.12.0 core, but not earlier.
@@ -15,6 +15,7 @@ our @EXPORT_OK = qw(
                      is_prime
                      is_prob_prime
                      is_provable_prime
+                     is_aks_prime
                      is_strong_pseudoprime
                      is_strong_lucas_pseudoprime
                      primes
@@ -41,11 +42,15 @@ BEGIN {
   eval {
     require XSLoader;
     XSLoader::load(__PACKAGE__, $Math::Prime::Util::GMP::VERSION);
+    _GMP_init();
     1;
   } or do {
     die $@;
   }
 }
+END {
+  _GMP_destroy();
+}
 
 sub _validate_positive_integer {
   my($n, $min, $max) = @_;
@@ -125,7 +130,7 @@ Math::Prime::Util::GMP - Utilities related to prime numbers and factoring, using
 
 =head1 VERSION
 
-Version 0.04
+Version 0.05
 
 
 =head1 SYNOPSIS
@@ -290,6 +295,16 @@ of the BPSW primality test (the Miller-Rabin strong pseudoprime test with
 base 2 being the other half).
 
 
+=head2 is_aks_prime
+
+  say "$n is definitely prime" if is_aks_prime($n);
+
+Takes a positive number as input, and returns 1 if the input passes the
+Agrawal-Kayal-Saxena (AKS) primality test.  This is a deterministic
+unconditional primality test which runs in polynomial time for general input.
+In practice, the BLS75 method used by is_provable_prime is much faster.
+
+
 =head2 primes
 
   my $aref1 = primes( 1_000_000 );
